/* rproc.c - a tiny SRIF file request processor                               */
/*                                                                            */
/* (c) 2004-2012 Alex Woick (2:244/1351) (alex@wombaz.de)                     */
/* -------------------------------------------------------------------------- */

#define VERSION "1.00"

#define _GNU_SOURCE
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include <fnmatch.h>
#include <getopt.h>
#include <limits.h>
#include <unistd.h>
#include <fcntl.h>
#include <utime.h>
#include <stdint.h>

#ifndef O_BINARY
#define O_BINARY 0
#endif

typedef unsigned char uchar;

/* define these if stdint.h is not available */
/*
typedef unsigned char  uint8_t;
typedef unsigned short uint16_t;
typedef unsigned long  uint32_t;
*/


#define DESCBUF          16384
#define LINEBUF          2048
#define BLOCK_MEMIDX     1000
#define BLOCK_HEAP       8192
#define LOGBUF           2048
#define COPYBUF          4096

#define PWDLEN           64

#define OKENTRY_FILESPEC 0
#define OKENTRY_MAGIC    1
#define OKENTRY_CMD      2
#define OKENTRY_NOTFOUND 3
#define OKENTRYMASK      0x0f
#define OKENTRY_COPY     0x80

#define OKFILE_COPY        0x01
#define OKFILE_SUBDIR_DESC 0x02
#define OKFILE_SUBDIR      0x04
#define OKFILE_LATEST      0x08

#define DIRSEP           '/'
#define DIRSEPSTR        "/"
#define WC_ALLFILES      "*"

#define RS_NORM          0
#define RS_LDESC         1
#define COMMENTCH        "\t -&:;*?'\"`[|]()$<=>{}^#\\%~"
#define LDESC_START      "          "

#define DB_FILE_NAME "file_name.idx"
#define DB_FILE_PWD  "file_pwd.idx"
#define DB_FILE_PATH "file_path.idx"
#define DB_FILE_DESC "file_desc.idx"
#define DB_FILE_INFO "file_info.idx"
#define DB_FILE_IDX  "file_idx.idx"


#define DB_CREATE        "w+b"
#define DB_OPENREADONLY  "rb"
#define DB_OPENREADWRITE "r+b"

/* ----------------------------------------------------------------------------
heap: A heap contains 0-terminated strings
  read string from heap:   heap_read(ulong ofs)
  write string to heap:    heap_write(char *s)
  remove string from heap: heap_delete(ulong ofs)

files:
  search file name: ptr = file_findfirst(char *s)
                          index_findnext(ptr)
----------------------------------------------------------------------------- */

struct _okfile {
  ushort level;
  char *name;
};


struct _desc {
  char *name;
  char *path;
  char *desc;
  struct _desc *next;
};

struct _db_memidx {
  char *pName;       /* filename */
  long ofsInfo;      /* offset into file_info.idx */
};


struct _heap {

  char *pData;
  ulong ulAllocated;
  ulong ulFree;
  long  ofsBase;
};


struct _db_info {
  long   ofsName;
  long   ofsPwd;
  long   ofsPath;
  long   ofsDesc;
  uchar  ucEntryType;
  ushort usLevel;
};


struct _db_userinfo {

  ushort level;
  ulong  maxfiles;
  ulong  maxbytes;
  ulong  maxtime;        /* seconds */
  char   *template;

};


struct _db {

  FILE *pNameIdx;  /* file is a heap of 0-terminated strings */
  FILE *pPwdIdx;   /* file is a heap of 0-terminated strings */
  FILE *pPathIdx;  /* file is a heap of 0-terminated strings */
  FILE *pDescIdx;  /* file is a heap of 0-terminated strings */
  FILE *pInfoIdx;  /* file has struct _db_info records       */
  FILE *pIdxIdx;   /* file is an array of long offsets to pInfoIdx */

  struct _db_memidx *memidx;
  ulong ulFreeFileIdx;
  ulong ulNumFileIdx;

  struct _heap HeapName;
  struct _heap HeapPwd;
  struct _heap HeapPath;
  struct _heap HeapDesc;

  ulong idxitems;   /* number of files in the index */

};


struct _fidonet_addr {

  ushort zone;
  ushort net;
  ushort node;
  ushort point;

};


struct _fidonet_addrlist {
  struct _fidonet_addr     addr;
  struct _fidonet_addrlist *next;
};


#define MSGPRIVATE 0x0001
#define MSGCRASH   0x0002
#define MSGREAD    0x0004
#define MSGSENT    0x0008
#define MSGFILE    0x0010
#define MSGFWD     0x0020
#define MSGORPHAN  0x0040
#define MSGKILL    0x0080
#define MSGLOCAL   0x0100
#define MSGHOLD    0x0200
#define MSGXX2     0x0400
#define MSGFRQ     0x0800
#define MSGRRQ     0x1000
#define MSGCPT     0x2000
#define MSGARQ     0x4000
#define MSGURQ     0x8000


struct _pkthdr {
  /* addresses */
  struct _fidonet_addr destAddr;
  struct _fidonet_addr origAddr;

  /* product codes */
  uchar hiProductCode;
  uchar loProductCode;
  uchar majorProductRev;
  uchar minorProductRev;

  /* date */
  time_t created;

  ushort capabilityWord;

  ulong prodData;

  char password[9]; /* password + \0 */
};


struct _msghdr {
  /* addresses */
  struct _fidonet_addr destAddr;
  struct _fidonet_addr origAddr;

  ushort attribute;
  char datetime[22];
  time_t timestamp;

  char toUserName[36];
  char fromUserName[36];
  char subject[72];
};





struct _cfg {
  char *reqidxdir;
  ushort level;
  struct _okfile *okfile;
  unsigned int numokfiles;
  char **mapfrom;
  char **mapto;
  unsigned int nummap;

  int got_unlisted;
  int got_listed;
  int got_protected;

  struct _db_userinfo unlisted;  /* userinfo for unlisted systems */
  struct _db_userinfo listed;    /* userinfo for listed systems */
  struct _db_userinfo protected; /* userinfo for protected systems */

  struct _fidonet_addrlist *myaddr;  /* my addresses */
  struct _fidonet_addr *myaddrmatch; /* my matching address */

  /* info from srif file */
  char *srif_sysop;                    /* name of remote sysop */
  struct _fidonet_addrlist *srif_aka;  /* address of remote sysop */
  struct _fidonet_addr *srif_akamatch; /* his matching address */
  long srif_baud;                      /* baudrate */
  long srif_time;                      /* time to next non-request event in minutes */
  char *srif_requestlist;
  char *srif_responselist;
  struct _db_userinfo *srif_userinfo;
  char srif_password[9];               /* session password, if protected */

  char *pktname;                       /* name of *.pkt response file   */
  FILE *pPkt;                          /* handle to *.pkt response file */
  FILE *pResponselist;                 /* handle to ASCII response file */

  long remaining_seconds; /* remaining available seconds for requested files */
  long remaining_files;   /* remaining available number of files */
  long remaining_bytes;   /* remaining available number of bytes */

};


typedef int (*CB_FINDFILE)(void *pData, const char *name,
                           const char *path, const char *desc, char type);


char *skipws(char *s)
{
  while (*s && isspace(*s)) ++s;
  return s;
}

char *skipnonws(char *s)
{
  while (*s && !isspace(*s)) ++s;
  return s;
}


char *xstrcat(char *s1, const char *s2)
{
  s1 = (char *) realloc(s1, strlen(s1) + strlen(s2) + 1);
  if (!s1)
    return NULL;

  strcat(s1, s2);
  return s1;
}


char *xstrntcat2(char *s1, const char *s2, const char *s3, size_t maxlen)
{
  size_t len = strlen(s1) + strlen(s2) + strlen(s3) + 1;
  s1 = (char *) realloc(s1, len);
  if (!s1)
    return NULL;

  strcat(s1, s2);
  strcat(s1, s3);
  if (len > maxlen) {
    s1[maxlen-1] = '\0';
    s1 = (char *) realloc(s1, maxlen);
  }
  return s1;
}


char *strntcpy(char *dest, const char *src, size_t n)
{
  strncpy(dest, src, n);
  dest[n-1] = '\0';
  return dest;
}


/* strip trailing newline-like chars */
void chomp(char *s)
{
  int l;
  while ((l = strlen(s)) != 0 && iscntrl(s[l-1])) {
    s[l-1] = '\0';
  }
}

char *loglevel = "XEWIT";
char *logfile = NULL;
int logquiet = 0;
char *loglevelerr = "XE";

void log(char lvl, const char *fmt, ...)
{
  char buf[LOGBUF];
  va_list ap;
  time_t t;
  size_t len;
  struct tm *tm;

  if (!strchr(loglevel, lvl))
    return;

  t = time(NULL);
  tm = localtime(&t);
  len = strftime(buf, LOGBUF, "%d.%m.%Y %T ", tm);
  va_start(ap, fmt);
  vsnprintf(buf+len, LOGBUF-len, fmt, ap);
  va_end(ap);

  if (!logquiet)
    puts(buf);

  if (logfile) {
    FILE *f = fopen(logfile, "a");
    if (f) {
      fputs(buf, f);
      putc('\n', f);
      fclose(f);
      return;
    }
  }

  if (logquiet && strchr(loglevelerr, lvl)) {
    fputs(buf, stderr);
    putc('\n', stderr);
  }
}


void copy_userinfo(struct _db_userinfo *pDest, struct _db_userinfo *pSrc)
{
  *pDest = *pSrc;
  if (pSrc->template)
    pDest->template = strdup(pSrc->template);
}


int fidonet_addrmatch(struct _fidonet_addr *a1, struct _fidonet_addr *a2)
{
  if (a1->zone != a2->zone) return 0;
  if (a1->net != a2->net) return 1;
  if (a1->node != a2->node) return 2;
  if (a1->point != a2->point) return 3;
  return 4;
}


struct _fidonet_addr *fidonet_addr(struct _fidonet_addr *a,
                  ushort z, ushort n, ushort f, ushort p)
{
  a->zone = z;
  a->net = n;
  a->node = f;
  a->point = p;
  return a;
}


char *fidonet_addr2str3d(char *s, struct _fidonet_addr *a)
{
  sprintf(s, "%hu:%hu/%hu", a->zone, a->net, a->node);
  return s;
}


char *fidonet_addr2str4d(char *s, struct _fidonet_addr *a)
{
  sprintf(s, "%hu:%hu/%hu.%hu", a->zone, a->net, a->node, a->point);
  return s;
}


char *fidonet_addr2str(char *s, struct _fidonet_addr *a)
{
  return a->point ? fidonet_addr2str4d(s, a) : fidonet_addr2str3d(s, a);
}


struct _fidonet_addr *fidonet_str2addr(char *s, struct _fidonet_addr *a)
{
  a->zone = a->net = a->node = a->point = 0;

  if (a && s) {
    char buf[24];
    char *p, *p1;

    p = strntcpy(buf, s, 24);
    p1 = strchr(p, '.');
    if (p1) {
      *(p1++) = '\0';
      a->point = (unsigned short) atol(p1);
    }
    p1 = strchr(p, '/');
    if (p1) {
      *(p1++) = '\0';
      a->node = (unsigned short) atol(p1);
    }
    p1 = strchr(p, ':');
    if (p1) {
      *(p1++) = '\0';
      a->net = (unsigned short) atol(p1);
      a->zone = (unsigned short) atol(p);
    } else
      a->net = (unsigned short) atol(p);
  }
  return a;
}

int fidonet_akamatch(struct _fidonet_addrlist *orig,
                     struct _fidonet_addrlist *dest,
                     struct _fidonet_addr **o,
                     struct _fidonet_addr **d)
{
  int longest = 0;
  struct _fidonet_addrlist *p, *q;

  *o = *d = NULL;
  for (p = orig; p; p = p->next)
  {
    for (q = dest; q; q = q->next)
    {
      int result = fidonet_addrmatch(&p->addr, &q->addr);
      if (result > longest) {
        longest = result;
        *o = &p->addr;
        *d = &q->addr;
      }
    }
  }
  if (longest == 0 && orig && dest) {
    *o = &orig->addr;
    *d = &dest->addr;
  }

  return longest;
}



int file_copy(const char *from, const char *to, int overwrite)
{
  char *buf;
  ssize_t bytesread;
  int s, d;
  struct stat st;
  struct utimbuf ut;

  if (stat(from, &st)) return -1; /* file does not exist */
  ut.actime = st.st_atime;
  ut.modtime = st.st_mtime;

  if (!(buf = (char *)malloc(COPYBUF))) return -1;

  if ((s = open(from, O_RDONLY|O_BINARY)) == -1) {
    free(buf);
    return -1;
  }

  if ((d = open(to, (overwrite ? 0 : O_EXCL)|O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,
                S_IREAD|S_IWRITE)) == -1) {
    close(s);
    free(buf);
    return -1;
  }

  while ((bytesread = read(s, buf, COPYBUF)) > 0) {
    if (write(d, buf, bytesread) != bytesread) {
      bytesread = -1;
      break;
    }
  }

  free(buf);
  close(d);
  close(s);

  if (bytesread == -1)
    unlink(to);

  utime(to, &ut);

  return bytesread; /* usually 0 */
}


char *get_tempdir(void)
{
  char *s;

  if ((s = getenv("TMP")) != NULL) return s;
  if ((s = getenv("TEMP")) != NULL) return s;
  if ((s = getenv("TMPDIR")) != NULL) return s;
#ifdef __unix__
  return DIRSEPSTR "tmp";
#else
  return "c:\\temp";
#endif
}


/* get zero-terminated string from f, maximum size n */
/* string is guaranteed to be zero-terminated */
size_t fgetszstr(unsigned char *s, size_t n, FILE *f)
{
  int ch;
  size_t i;
  size_t maxchars = n - 1;

  for (i=0; i<maxchars; i++) {
    if ((ch = getc(f)) == EOF) {
      s[i] = '\0';
      return i+1;
    } else {
      if (!(s[i] = (uint8_t) ch))
        return i+1;
    }
  }
  s[i] = '\0';
  return n;
}


uint16_t fgetUINT16(FILE *f)
{
  /* this is correct, since short-cut evaluation is mandatory in C */
  return ((uint16_t)(uint8_t)getc(f)) | (((uint16_t)(uint8_t)getc(f)) << 8);
}

uint16_t fgetUINT32(FILE *f)
{
  /* this is correct, since short-cut evaluation is mandatory in C */
  return ((uint32_t)(uint8_t)getc(f)) | (((uint32_t)(uint8_t)getc(f)) << 8) |
         (((uint32_t)(uint8_t)getc(f)) << 16) | (((uint32_t)(uint8_t)getc(f)) << 24);
}




char *put_UINT8(char *p, uint8_t data)
{
  *(p++) = (char)data;
  return p;
}


char *put_UINT16(char *p, uint16_t data)
{
  *(p++) = (char)data;               /*  write low Byte */
  *(p++) = (char)(data >> 8);        /*  write high Byte */
  return p;
}


char *put_UINT32(char *p, uint32_t data)
{
  *(p++) = (char)data;
  *(p++) = (char)(data>>8);
  *(p++) = (char)(data>>16);
  *(p++) = (char)(data>>24);
  return p;
}



char *get_UINT32(char *p, uint32_t *data)
{
  *data =  (uint32_t) (uint8_t)*(p++);
  *data |= ((uint32_t) (uint8_t)*(p++)) << 8;
  *data |= ((uint32_t) (uint8_t)*(p++)) << 16;
  *data |= ((uint32_t) (uint8_t)*(p++)) << 24;
  return p;
}


char *get_UINT16(char *p, uint16_t *data)
{
  *data = (uint16_t) (uint8_t)*(p++);
  *data |= ((uint16_t) (uint8_t)*(p++)) << 8;
  return p;
}

char *get_UINT8(char *p, uint8_t *data)
{
  *data = (uint8_t)*(p++);
  return p;
}


char *get_fidonet_addr(char *p, struct _fidonet_addr *a)
{
  p = get_UINT16(p, &a->zone);
  p = get_UINT16(p, &a->net);
  p = get_UINT16(p, &a->node);
  p = get_UINT16(p, &a->point);
  return p;
}

char *put_fidonet_addr(char *p, struct _fidonet_addr *a)
{
  p = put_UINT16(p, (uint16_t) a->zone);
  p = put_UINT16(p, (uint16_t) a->net);
  p = put_UINT16(p, (uint16_t) a->node);
  p = put_UINT16(p, (uint16_t) a->point);
  return p;
}


char *get_userinfo(char *p, struct _db_userinfo *dbui)
{
  uint16_t ui16;
  uint32_t ui32;
  ushort strsize;

  p = get_UINT16(p, &ui16); dbui->level =    (ushort) ui16;
  p = get_UINT32(p, &ui32); dbui->maxfiles = (ulong) ui32;
  p = get_UINT32(p, &ui32); dbui->maxbytes = (ulong) ui32;
  p = get_UINT32(p, &ui32); dbui->maxtime  = (ulong) ui32;
  p = get_UINT16(p, &ui16); strsize =        (ushort) ui16;
  if (!(dbui->template = (char *) malloc(strsize))) {
    log('E', "error: unable to allocate memory for database header processing");
    exit(4);
  }
  strntcpy(dbui->template, p, strsize);
  return p + strsize;
}


char *put_userinfo(char *p, struct _db_userinfo *dbui)
{
  p = put_UINT16(p, (uint16_t) dbui->level);
  p = put_UINT32(p, (uint32_t) dbui->maxfiles);
  p = put_UINT32(p, (uint32_t) dbui->maxbytes);
  p = put_UINT32(p, (uint32_t) dbui->maxtime);
  if (dbui->template) {
    int strsize = strlen(dbui->template)+1;
    p = put_UINT16(p, (uint16_t) strsize);
    strcpy(p, dbui->template);
    p += strsize;
  } else {
    p = put_UINT16(p, 1);
    *(p++) = '\0';
  }
  return p;
}


ulong GenMsgId(void)
{
  static ushort counter = 0;
  if (!++counter) sleep(1);
  return (ulong)time(NULL) ^ (((ulong) (ushort) getpid()) | (((ulong)counter) << 16));
}


long pkt_writehdr(FILE *pkt, struct _pkthdr *hdr)
{
  struct tm *tm;
  char buf[58];
  char *p;
  long written;

  log('D', "pkt_writehdr begin");

  p = put_UINT16(buf, (uint16_t)hdr->origAddr.node);
  p = put_UINT16(p,   (uint16_t)hdr->destAddr.node);

  /*  write time */
  tm = localtime(&(hdr->created));
  p = put_UINT16(p, (uint16_t)(tm->tm_year + 1900));
  p = put_UINT16(p, (uint16_t) tm->tm_mon);
  p = put_UINT16(p, (uint16_t) tm->tm_mday);
  p = put_UINT16(p, (uint16_t) tm->tm_hour);
  p = put_UINT16(p, (uint16_t) tm->tm_min);
  p = put_UINT16(p, (uint16_t) tm->tm_sec);

  /*  write unused baud field */
  p = put_UINT16(p, 0);

  /*  write pktver == 2 */
  p = put_UINT16(p, 2);

  /*  write net info */
  p = put_UINT16(p, (uint16_t) hdr->origAddr.net);
  p = put_UINT16(p, (uint16_t) hdr->destAddr.net);

  p = put_UINT8(p, hdr->loProductCode);   /*  put lowByte of Prod-Id */
  p = put_UINT8(p, hdr->majorProductRev); /*  put major version number */

  /*  write PKT pwd max 8 char; fill the rest with \0 */
  memset(p, 0, 8);
  strncpy(p, hdr->password, 8); /* here is strncpy correct! */
  p += 8;

  /*  write qzone info */
  p = put_UINT16(p, (uint16_t) hdr->origAddr.zone);
  p = put_UINT16(p, (uint16_t) hdr->destAddr.zone);

  p = put_UINT16(p, 0); /*  filler */

  /*  write byte swapped capability Word */
  p = put_UINT8(p, (uint8_t)(hdr->capabilityWord >> 8));
  p = put_UINT8(p, (uint8_t)(hdr->capabilityWord & 0x00ff));

  p = put_UINT8(p, hdr->hiProductCode);      /*  put hiByte of Prod-Id */
  p = put_UINT8(p, hdr->minorProductRev);    /*  put minor version number */

  p = put_UINT16(p, hdr->capabilityWord);

  p = put_UINT16(p, (uint16_t) hdr->origAddr.zone);
  p = put_UINT16(p, (uint16_t) hdr->destAddr.zone);

  p = put_UINT16(p, (uint16_t) hdr->origAddr.point);
  p = put_UINT16(p, (uint16_t) hdr->destAddr.point);

  /* product specific data, usually 0 or serial number */
  p = put_UINT32(p, hdr->prodData);

  written = fwrite(buf, 1, 58, pkt);
  log('D', "pkt_writehdr end, written=%ld", written);
  return written;
}


long pkt_writemsghdr(FILE *pkt, struct _msghdr *hdr)
{
  char buf[34+36+36+72];
  char *p;
  static char months_ab[][4] =
  {
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec"
  };

  p = put_UINT16(buf, 2);     /* type */

  p = put_UINT16(p, (uint16_t)hdr->origAddr.node);
  p = put_UINT16(p, (uint16_t)hdr->destAddr.node);
  p = put_UINT16(p, (uint16_t)hdr->origAddr.net);
  p = put_UINT16(p, (uint16_t)hdr->destAddr.net);

  p = put_UINT16(p, hdr->attribute);

  p = put_UINT16(p, 0); /* cost */

  if (!*hdr->datetime) {
    struct tm *tm = localtime(&hdr->timestamp);
    snprintf(hdr->datetime, 20, "%02d %3s %02d  %02d:%02d:%02d",
      tm->tm_mday, months_ab[tm->tm_mon], tm->tm_year%100,
      tm->tm_hour, tm->tm_min, tm->tm_sec);
  }
  strntcpy(p, hdr->datetime, 20);
  p += 20;
  strntcpy(p, hdr->toUserName, 36);
  p += strlen(p)+1;
  strntcpy(p, hdr->fromUserName, 36);
  p += strlen(p)+1;
  strntcpy(p, hdr->subject, 72);
  p += strlen(p)+1;

  return fwrite(buf, 1, p-buf, pkt);
}


/*
write the following kludges:
INTL zzzzz:nnnnn/fffff zzzzz:nnnnn/fffff
FMPT ppppp
TOPT ppppp
MSGID zzzzz:nnnnn/fffff.ppppp ssssssss
FLAGS NPD

*/

#define CTRLBUF (42+12+12+40+11+1)
long pkt_writemsgkludges(FILE *pkt, struct _msghdr *msghdr)
{
  char a1[24];
  char a2[24];
  char buf[CTRLBUF];
  char *p;

  p = buf + snprintf(buf, 43, "\001INTL %s %s\r",
                           fidonet_addr2str3d(a1, &msghdr->destAddr),
                           fidonet_addr2str3d(a2, &msghdr->origAddr));
  if (msghdr->origAddr.point)
    p += snprintf(p, 13, "\001FMPT %u\r", msghdr->origAddr.point);
  if (msghdr->destAddr.point)
    p += snprintf(p, 13, "\001TOPT %u\r", msghdr->destAddr.point);

  p += snprintf(p, 41, "\001MSGID: %s %08lx\r",
         fidonet_addr2str(a1, &msghdr->origAddr), GenMsgId());

  p += snprintf(p, 12, "\001FLAGS NPD\r");

  return fwrite(buf, 1, p-buf, pkt);
}


long pkt_writemsg(FILE *pkt, struct _msghdr *msghdr, char *msgtxt)
{
  long bytes;

  bytes =  pkt_writemsghdr(pkt, msghdr);
  bytes += pkt_writemsgkludges(pkt, msghdr);
  if (msgtxt)
    bytes += fwrite(msgtxt, 1, strlen(msgtxt)+1, pkt);

  return bytes;
}


long pkt_close(FILE *pkt)
{
  int rc;

  rc = fwrite("\000\000", 1, 2, pkt);
  fclose(pkt);
  return rc;
}




/* ------------------------------------------------------------------------- */


/* init/recycle existing heap */
struct _heap *heap_init(struct _heap *heap)
{
  if (heap->pData) {
    heap->ulFree += heap->ulAllocated;
    heap->ulAllocated = 0;
  } else {
    heap->ulAllocated = heap->ulFree = 0;
  }
  return heap;
}


/* initial creation of an empty heap */
struct _heap *heap_create(struct _heap *heap)
{
  if (!heap) {
    if (!(heap = (struct _heap *) malloc(sizeof(struct _heap)))) {
      log('E', "error: unable to allocate string memory");
      exit(4);
    }
  }
  heap->ofsBase = 0;
  heap->pData = NULL;
  heap->ulAllocated = heap->ulFree = 0;
  return heap;
}


long heap_addstr(struct _heap *heap, const char *s)
{
  ulong pos;
  size_t needed;

/*  log('D', "heap_addstr s='%s'", s); */

  if (!s)
    return -1;

  needed = strlen(s) + 1;

  while (needed > heap->ulFree) {
    heap->pData = (char *) realloc(heap->pData, heap->ulAllocated + heap->ulFree + BLOCK_HEAP);
    if (!heap->pData) {
      log('E', "error: unable to allocate string memory");
      exit(4);
    }
    heap->ulFree += BLOCK_HEAP;
  }
  pos = heap->ulAllocated;
  strcpy(heap->pData + pos, s);
  heap->ulFree -= needed;
  heap->ulAllocated += needed;

/*  log('D', "heap_addstr: new ofs=%lu", heap->ofsBase + pos); */

  return heap->ofsBase + pos;
}


long heap_findstr(struct _heap *heap, const char *s)
{
  char *p, *pEnd;

/*  log('D', "heap_findstr s='%s'", s); */

  if (s) {
    pEnd = heap->pData + heap->ulAllocated;

    for (p = heap->pData; p < pEnd; p += strlen(p) + 1) {
      if (!strcmp(p, s)) {
/*        log('D', "heap_findstr found at offset=%lu", heap->ofsBase + p - heap->pData); */
        return heap->ofsBase + p - heap->pData;
      }
    }
  }
  return -1;
}


void heap_writetofile(struct _heap *heap, FILE *pFile)
{
  size_t written;

/*  log('D', "heap_writetofile start, heap->ulAllocated=%lu", heap->ulAllocated); */

  if (heap->ulAllocated) {

    written = fwrite(heap->pData, 1, (long) heap->ulAllocated, pFile);
    if (written != heap->ulAllocated) {
      log('E', "error: unable to write to file (disk full?)");
      exit(5);
    }
    heap->ofsBase += written;
  }

/*  log('D', "heap_writetofile end, heap->ofsBase=%lu", heap->ofsBase); */
}


/* ------------------------------------------------------------------------- */



/* get heap string from file */
char *db_heapstr_get(FILE *pFile, long ofs, char *pBuf, int buflen)
{
  char *p = pBuf;
  char *pEnd = pBuf + buflen - 1;
  int ch;

  if (ofs != -1) {
    fseek(pFile, ofs, SEEK_SET);
    while (p < pEnd && (ch=getc(pFile)) != EOF && ch) {
      *(p++) = (char) ch;
    }
  }
  *p = '\0';

  return pBuf;
}


/* get offset to info record from file_idx.idx */
long db_idx_get(struct _db *db, long idx)
{
  long numread;
  uint32_t ofs;
  char buf[sizeof(uint32_t)];

/*  log('D', "db_idx_get: get index record %ld from offset 0x%08lx",
            idx, idx * sizeof(long)); */

  fseek(db->pIdxIdx, idx * sizeof(uint32_t), SEEK_SET);
  numread = fread(buf, sizeof(uint32_t), 1, db->pIdxIdx);
  if (!numread || feof(db->pIdxIdx)) {
    log('E', "db_idx_get: unexpected end of file in '%s'", DB_FILE_IDX);
    return -1;
  }
  get_UINT32(buf, &ofs);
  return (long) ofs;
}


/* get item from file_info.idx */
int db_info_get(struct _db *db, long ofs, struct _db_info *pInfo)
{
  long numread;
  char buf[19];
  char *p;
  uint32_t ui32;
  uint16_t ui16;
  uint8_t ui8;

/*  log('D', "db_info_get: get info record from offset 0x%08lx", ofs); */

  /* seek to info block and read info block */
  fseek(db->pInfoIdx, ofs, SEEK_SET);
  numread = fread(buf, 1, 19, db->pInfoIdx);
  if (!numread) {
    log('E', "db_info_get: unexpected end of file in '%s'", DB_FILE_INFO);
    return -1;
  }
  p = get_UINT32(buf, &ui32); pInfo->ofsName = (int32_t) ui32;
  p = get_UINT32(p,   &ui32); pInfo->ofsPwd  = (int32_t) ui32;
  p = get_UINT32(p,   &ui32); pInfo->ofsPath = (int32_t) ui32;
  p = get_UINT32(p,   &ui32); pInfo->ofsDesc = (int32_t) ui32;
  p = get_UINT8(p,    &ui8);  pInfo->ucEntryType = (uchar) ui8;
  p = get_UINT16(p,   &ui16); pInfo->usLevel = (ushort) ui16;

  return 0;
}


/* get item from file_info.idx by index into file_idx.idx */
int db_idx_info_get(struct _db *db, long idx, struct _db_info *pInfo)
{
  long ofsInfo;

  /* get offset from index */
  ofsInfo = db_idx_get(db, idx);
  if (ofsInfo == -1)
    return -1;

  /* get record from data file */
  return db_info_get(db, ofsInfo, pInfo);
}


/* get filename of file indexed by idx in the database index */
char *db_idx_getfilename(struct _db *db, long idx, char *pBuf, int buflen)
{
  struct _db_info info;

  if (db_idx_info_get(db, idx, &info) == -1)
    return NULL;

  return db_heapstr_get(db->pNameIdx, info.ofsName, pBuf, buflen);
}


/* compare the item at position idx (from the index) with key */
int db_compare_filename(struct _db *db, long idx, const char *key, size_t n)
{
  int result;
  char buf[PATH_MAX];

  db_idx_getfilename(db, idx, buf, PATH_MAX);

  result = strncasecmp(buf, key, n);
/*  log('D', "db_compare_filename s='%s', key='%s', n=%d, result=%d", s, key, n, result); */
  return result;
}



/* return index into file_idx.idx for first filename matching key */
long db_idx_search(struct _db *db, const char *key, size_t n)
{
  long low = -1;
  long high = db->idxitems;
  long probe;

  while (high - low > 1)
  {
    probe = (high + low) >> 1;
    if (db_compare_filename(db, probe, key, n) < 0)
      low = probe;
    else
      high = probe;
  }
  if (high == db->idxitems || db_compare_filename(db, high, key, n) != 0)
    return -1;
  else
    return high;
}



/* open one database file */
FILE *db_open_file(const char *path, const char *name, const char *openmode)
{
  FILE *pDbFile;
  char buf[PATH_MAX];

  log('D', "db_open_file: path='%s', name='%s', openmode='%s'",
           path, name, openmode);

  if (strlen(path) + strlen(name) + 1 > PATH_MAX) {
    log('E', "error: path to long: '%s/%s'", path, name);
    exit(2);
  }
  strcpy(buf, path);
  strcat(buf, DIRSEPSTR);
  strcat(buf, name);

  if (!(pDbFile = fopen(buf, openmode))) {
    log('E', "error: unable to open file '%s'", buf);
    exit(3);
  }
  return pDbFile;
}


/* open database */
struct _db *db_open(const char *path, const char *openmode)
{
  struct _db *db;

  log('D', "db_open path='%s', openmode='%s'", path, openmode);

  if (!(db = (struct _db *) malloc(sizeof(struct _db)))) {
    log('E', "error: unable to allocate memory for database handling");
    exit(4);
  }
  db->pNameIdx = db_open_file(path, DB_FILE_NAME, openmode);
  db->pPwdIdx  = db_open_file(path, DB_FILE_PWD,  openmode);
  db->pPathIdx = db_open_file(path, DB_FILE_PATH, openmode);
  db->pDescIdx = db_open_file(path, DB_FILE_DESC, openmode);
  db->pInfoIdx = db_open_file(path, DB_FILE_INFO, openmode);
  db->pIdxIdx  = db_open_file(path, DB_FILE_IDX,  openmode);

  db->memidx = NULL;
  db->ulFreeFileIdx = 0;
  db->ulNumFileIdx = 0;

  heap_create(&db->HeapName);
  heap_create(&db->HeapPwd);
  heap_create(&db->HeapPath);
  heap_create(&db->HeapDesc);

  fseek(db->pIdxIdx, 0, SEEK_END);
  db->idxitems = ftell(db->pIdxIdx) / sizeof(uint32_t);

  return db;
}


struct _db *db_open_read(struct _cfg *cfg)
{
  char buf4[sizeof(uint32_t)];
  char *buf;
  char *p;
  struct _db *db = db_open(cfg->reqidxdir, DB_OPENREADONLY);
  size_t numread;
  long hdrsize;
  uint16_t numaddrs;
  int i;
  uint32_t ui32;

  numread = fread(buf4, sizeof(uint32_t), 1, db->pInfoIdx);
  if (!numread || feof(db->pInfoIdx)) {
    log('E', "db_open_read: unexpeced end of file in '%s'", DB_FILE_INFO);
    exit(5);
  }
  get_UINT32(buf4, &ui32); hdrsize = (long) ui32;
  log('D', "db_open_read: hdrsize=%ld", hdrsize);
  if (!(buf = (char *) malloc(hdrsize))) {
    log('E', "db_open_read: unable to allocate buffer for header processing");
    exit(4);
  }

  numread = fread(buf, 1, hdrsize, db->pInfoIdx);
  if (!numread || feof(db->pInfoIdx)) {
    log('E', "db_open_read: unexpeced end of file in '%s'", DB_FILE_INFO);
    exit(5);
  }
  p = get_userinfo(buf, &cfg->unlisted);
  p = get_userinfo(p,   &cfg->listed);
  p = get_userinfo(p,   &cfg->protected);

  /* read my addresses */
  p = get_UINT16(p, &numaddrs);

  for (i=0; i<numaddrs; i++) {
    struct _fidonet_addrlist *a = (struct _fidonet_addrlist *) malloc(sizeof(struct _fidonet_addrlist));
    a->next = cfg->myaddr;
    cfg->myaddr = a;
    p = get_fidonet_addr(p, &a->addr);
  }

  free(buf);

  return db;
}

/* create database */
struct _db *db_create(struct _cfg *cfg)
{
  struct _db *db;
  char *buf, *p, *p1;
  size_t written;
  struct _fidonet_addrlist *a;
  uint16_t numaddrs = 0;

  log('D', "db_create, reqidxdir='%s'", cfg->reqidxdir);
  db = db_open(cfg->reqidxdir, DB_CREATE);

  /* count my addresses */
  for (a = cfg->myaddr; a; a = a->next)
    ++numaddrs;

  /* allocate buffer for header creation */
  if (!(buf = (char *) malloc(2 + numaddrs * sizeof(struct _fidonet_addr) + 3 * (sizeof(struct _db_userinfo) + PATH_MAX)))) {
    log('E', "db_create: unable to allocate buffer for header processing");
    exit(4);
  }

  /* configuration goes into a database header */

  /* start with dummy size, to be filled in later */
  p1 = put_UINT32(buf, 0L); 

  /* write user limits info */
  p = put_userinfo(p1, &cfg->unlisted);
  p = put_userinfo(p, &cfg->listed);
  p = put_userinfo(p, &cfg->protected);

  /* write my addresses */
  p = put_UINT16(p, numaddrs);
  for (a = cfg->myaddr; a; a = a->next)
    p = put_fidonet_addr(p, &a->addr);

  put_UINT32(buf, p - p1); /* this is the real header size */

  log('D', "db_create: write %d bytes header, PATH_MAX=%d", (p-buf), PATH_MAX);

  /* write to database */
  written = fwrite(buf, 1, p-buf, db->pInfoIdx);
  if (written != p-buf) {
    log('E', "error: unable to write config to %s (disk full?)", DB_FILE_INFO);
    exit(5);
  }

  return db;
}


/* close database */
int db_close(struct _db *db)
{
  unsigned long i;

  log('D', "db_close");

  fclose(db->pNameIdx);
  fclose(db->pPwdIdx);
  fclose(db->pPathIdx);
  fclose(db->pDescIdx);
  fclose(db->pInfoIdx);
  fclose(db->pIdxIdx);

  free(db->HeapName.pData);
  free(db->HeapPwd.pData);
  free(db->HeapPath.pData);
  free(db->HeapDesc.pData);

  if (db->memidx) {
    for (i=0; i<db->ulNumFileIdx; i++) {
      free(db->memidx[i].pName);
    }
    free(db->memidx);
  }
  free(db);
  return 0;
}


struct _db_memidx *db_memidx_new(struct _db *db)
{
  if (!db->ulFreeFileIdx) {
    db->memidx = (struct _db_memidx *) realloc(db->memidx,
                              sizeof(struct _db_memidx) * (db->ulNumFileIdx + BLOCK_MEMIDX));
    if (!db->memidx) {
      log('E', "error: unable to allocate more memory for file index");
      exit(4);
    }
    db->ulFreeFileIdx = BLOCK_MEMIDX;
  }
  --db->ulFreeFileIdx;
  return &db->memidx[db->ulNumFileIdx++];
}



/* add file to database */
void db_file_add(struct _db *db, const char *filename, const char *pwd,
             const char *filepath, const char *filedesc, uchar entrytype,
             ushort level)
{
/*
   filename to file_name.idx
   pwd to file_pwd.idx
   filepath to file_path.idx
   filedesc to file_desc.idx
   offsets to the above strings and entrytype to file_info.idx
   save offset into file_info.idx and corresponding filename in memory
*/
  struct _db_memidx *pMemIdx = db_memidx_new(db);
  struct _db_info info;
  char buf[sizeof(struct _db_info)];
  char bufpwd[PWDLEN];
  char *p;
  long written;

  log('D', "db_file_add: filename='%s', pwd='%s', filepath='%s', filedesc='%s'",
           filename, pwd, filepath, filedesc);

  /* save data in the heaps */
  info.ofsName = heap_addstr(&db->HeapName, filename);

  if (pwd) {
    strntcpy(bufpwd, pwd, PWDLEN);
    info.ofsPwd = heap_findstr(&db->HeapPwd, bufpwd);
    if (info.ofsPwd == -1)
      info.ofsPwd = heap_addstr(&db->HeapPwd, bufpwd);
  } else {
    info.ofsPwd = -1;
  }

  info.ofsPath = heap_findstr(&db->HeapPath, filepath);
  if (info.ofsPath == -1)
    info.ofsPath = heap_addstr(&db->HeapPath, filepath);

  info.ofsDesc = heap_addstr(&db->HeapDesc, filedesc);

  info.ucEntryType = entrytype;
  info.usLevel = level;

  /* save index info */
  pMemIdx->ofsInfo = ftell(db->pInfoIdx);
  pMemIdx->pName = strdup(filename);

  /* write info record to disk */
  p = put_UINT32(buf, info.ofsName);
  p = put_UINT32(p,   info.ofsPwd);
  p = put_UINT32(p,   info.ofsPath);
  p = put_UINT32(p,   info.ofsDesc);
  p = put_UINT8(p,    info.ucEntryType);
  p = put_UINT16(p,   info.usLevel);

  written = fwrite(buf, 1, p-buf, db->pInfoIdx);
  if (written != p-buf) {
    log('E', "error: unable to write to %s (disk full?)", DB_FILE_INFO);
    exit(5);
  }
}


/* begin group of files */
int db_area_start(struct _db *db)
{

  log('D', "db_area_start");

  /* initialize all heaps */
  heap_init(&db->HeapName);
  heap_init(&db->HeapPwd);
  heap_init(&db->HeapPath);
  heap_init(&db->HeapDesc);

  return 0;
}


/* end group of files (write to disk) */
int db_area_end(struct _db *db)
{

  log('D', "db_area_end");

  /* write the heaps to disk */
  heap_writetofile(&db->HeapName, db->pNameIdx);
  heap_writetofile(&db->HeapPwd,  db->pPwdIdx);
  heap_writetofile(&db->HeapPath, db->pPathIdx);
  heap_writetofile(&db->HeapDesc, db->pDescIdx);

  return 0;
}


int compare_memidx(const void *p1, const void *p2)
{
  struct _db_memidx *idx1 = (struct _db_memidx*)p1;
  struct _db_memidx *idx2 = (struct _db_memidx*)p2;

/*  log('D', "compare_memidx idx1=%p, idx2=%p", idx1, idx2); */

  return strcasecmp(idx1->pName, idx2->pName);
}


/* sort the main index */
void db_sort_idx(struct _db *db)
{
/*  log('D', "db_sort_idx start, db->ulNumFileIdx=%lu", db->ulNumFileIdx); */

  qsort(db->memidx, db->ulNumFileIdx, sizeof(*db->memidx), compare_memidx);

/*  log('D', "db_sort_idx end"); */
}


/* write the main index to disk */
int db_write_idx(struct _db *db)
{
  char *buf, *p;
  long bufsize;
  long items;
  long i;
  long written;

/*  log('D', "db_write_idx start"); */

  bufsize = BLOCK_MEMIDX * sizeof(uint32_t);
  if (!(buf = (char *) malloc(bufsize))) {
    log('E', "error: unable to allocate memory for index creation");
    exit(4);
  }

/*  log('D', "db_write_idx: going to write %lu items", db->ulNumFileIdx); */
  p = buf;
  items = 0;
  for (i=0; i<db->ulNumFileIdx; i++, items++) {
    if (items == BLOCK_MEMIDX) {
/*      log('D', "db_write_idx: write %lu items of %lu bytes", items, sizeof(long)); */
      written = fwrite(buf, sizeof(uint32_t), items, db->pIdxIdx);
      if (written != items) {
        log('E', "error: unable to write to %s (disk full?)", DB_FILE_IDX);
        exit(5);
      }
      items = 0;
      p = buf;
    }
    p = put_UINT32(p, db->memidx[i].ofsInfo);
  }
  if (items) {
    written = fwrite(buf, sizeof(uint32_t), items, db->pIdxIdx);
    if (written != items) {
      log('E', "error: unable to write to %s (disk full?)", DB_FILE_IDX);
      exit(5);
    }
  }

  free(buf);
/*  log('D', "db_write_idx end"); */
  return 0;
}

void dump_userinfo(struct _db_userinfo *pui)
{
  printf("    level=%hu\n"
         "    maxfiles=%lu\n"
         "    maxbytes=%lu\n"
         "    maxtime=%lu\n"
         "    template=%s\n", pui->level, pui->maxfiles, pui->maxbytes,
                            pui->maxtime, pui->template);

}


int db_dump(struct _cfg *cfg)
{
  struct _db *db;
  char typedesc[80];
  struct _db_info info;
  long idx;
  char *filename = (char *) malloc(PATH_MAX);
  char *filepwd  = (char *) malloc(PWDLEN);
  char *filepath = (char *) malloc(PATH_MAX);
  char *filedesc = (char *) malloc(DESCBUF);

  if (!(filename && filepwd && filepath && filedesc)) {
    log('E', "db_dump: unable to allocate memory");
    exit(4);
  }

  printf("configuration from command line:\n"
         "  reqidxdir=%s\n", cfg->reqidxdir);
  for (idx=0; idx<cfg->numokfiles; idx++) {
    printf("  okfile=%s, level=%hu\n",
             cfg->okfile[idx].name, cfg->okfile[idx].level);
  }
  printf("  userinfo unlisted (got_unlisted=%d)\n", cfg->got_unlisted);
  dump_userinfo(&cfg->unlisted);
  printf("  userinfo listed (got_listed=%d)\n", cfg->got_listed);
  dump_userinfo(&cfg->listed);
  printf("  userinfo protected (got_protected=%d)\n", cfg->got_protected);
  dump_userinfo(&cfg->protected);


  db = db_open_read(cfg);

  printf("\nconfiguration from database:\n");
  printf("  userinfo unlisted\n");
  dump_userinfo(&cfg->unlisted);
  printf("  userinfo listed\n");
  dump_userinfo(&cfg->listed);
  printf("  userinfo protected\n");
  dump_userinfo(&cfg->protected);

  printf("\n");
  printf("---------- files ----------\n");

  for (idx = 0; idx < db->idxitems; idx++) {

    /* get file info entry */
    if (db_idx_info_get(db, idx, &info) != 0)
      break;

    /* read filename */
    db_heapstr_get(db->pNameIdx, info.ofsName, filename, PATH_MAX);

    /* read password */
    db_heapstr_get(db->pPwdIdx, info.ofsPwd, filepwd, PWDLEN);

    /* read file path */
    db_heapstr_get(db->pPathIdx, info.ofsPath, filepath, PATH_MAX);

    /* read file description */
    db_heapstr_get(db->pDescIdx, info.ofsDesc, filedesc, DESCBUF);

    /* print file info */
    switch (info.ucEntryType & OKENTRYMASK) {
      case OKENTRY_FILESPEC:
        strcpy(typedesc, "NORMAL FILE");
        break;
      case OKENTRY_MAGIC:
        strcpy(typedesc, "MAGIC");
        break;
      case OKENTRY_CMD:
        strcpy(typedesc, "EXTERNAL COMMAND");
        break;
      case OKENTRY_NOTFOUND:
        strcpy(typedesc, "NOT FOUND");
        break;
      default:
        strcpy(typedesc, "**ERROR**");
    }
    if (info.ucEntryType & OKENTRY_COPY) {
      strcat(typedesc, " - COPY");
    }

    printf("filename: '%s'\n"
           "    type: %s\n"
           "   level: %hu\n"
           "    path: '%s'\n"
           "    pwd:  '%s'\n"
           "    desc: '%s'\n",
                filename, typedesc, info.usLevel, filepath, filepwd, filedesc);
  }

  db_close(db);
  free(filename);
  free(filepwd);
  free(filepath);
  free(filedesc);
  return 0;
}


char * mappath(char *dst, const char *src, int dstsize, struct _cfg *cfg)
{
  unsigned int i;

/*  log('D', "mappath: src='%s', dstsize=%d", src, dstsize); */

  strntcpy(dst, src, dstsize);
  for (i=0; i<cfg->nummap; i++) {
    size_t fromlen = strlen(cfg->mapfrom[i]);
    size_t tolen = strlen(cfg->mapto[i]);
    char *p = dst;
    while (*p && strlen(dst) + tolen < dstsize + fromlen) {
      if (strncasecmp(p, cfg->mapfrom[i], fromlen) == 0) {
        if (fromlen != tolen)
          memmove(p+tolen, p+fromlen, strlen(p+fromlen)+1);
        memcpy(p, cfg->mapto[i], tolen);
        p += tolen;
      } else
        p++;
    }
  }
  log('D', "mappath: src='%s', dstsize=%d -> dst='%s'", src, dstsize, dst);

  return dst;
}




struct _desc *description_read(struct _cfg *cfg, const char *filename)
{
  struct _desc *list = NULL;
  struct _desc *d = NULL;
  FILE *pFile;
  char *pBuf, *line, *p;
  int len;                  /* line length without trailing \n */
  int state;
  char *fname;
  int fnlen;                /* length of scanned file names */
  char ch;
  char buf[PATH_MAX];

  log('D', "description_read start, filename='%s'", filename);

  pBuf = (char *) malloc(DESCBUF);
  if (!pBuf) {
    log('E', "error: unable to allocate memory for description file processing");
    return NULL;
  }

  if (!(pFile = fopen(filename, "rt")))
  {
    log('D', "description_read: unable to open description file '%s'", filename);
    return NULL;
  }

  state = RS_NORM;
  fname = NULL;
  while ((line = fgets(pBuf, DESCBUF, pFile)) != NULL) {
    chomp(line);

    /* line types:
       description: line is empty or doesn't start with a valid filename char
       filename:    line starts with a valid filename char
       long description: previous line was filename or long description */

    /* empty line: reset state */
    if (!(len = strlen(line))) {
      state = RS_NORM;
      continue;
    }

    /* do we have a comment? */
    ch = *line;
    if (ch < ' ' || strchr(COMMENTCH, ch)) {

      /* it is a comment: also a ldesc? */
      if (state == RS_LDESC && !strncmp(line, LDESC_START, strlen(LDESC_START))) {

        /* yes: ldesc. Append to current description */
        p = skipws(line);
/*        log('D', "description_read: found ldesc='%s'", p); */
        d->desc = xstrntcat2(d->desc, "\n", p, USHRT_MAX);
/*        log('D', "description_read: complete desc='%s'", d->desc); */

      } else {

        /* no: only normal comment. reset scanner. */
        state = RS_NORM;
/*        log('D', "description_read: skipping comment '%s'", line); */
      }
    } else {

      /* no comment, so we have a filename */
      log('D', "description_read: found file='%s'", line);
      if (list) {
        d->next = (struct _desc *) malloc(sizeof(struct _desc));
        d = d->next;
      } else
        d = list = (struct _desc *) malloc(sizeof(struct _desc));

      d->next = NULL;

      /* split path, filename and description */
      fnlen = strcspn(line, " \t");
      p = line + fnlen;
      if (*p)
        *(p++) = '\0';

      p = skipws(p);
      if (*p == '[') {      /* detect downloadcounter */
        char *p1 = p + 1;   /* skip opening brace */
        while (isdigit(*p1) || isspace(*p1))
          ++p1;             /* skip digits or spaces */
        if (*p1 == ']')
          p = p1+1;         /* skip closing brace */
      }
      d->desc = strdup(skipws(p));

      /* split path from filename */
      mappath(buf, line, PATH_MAX, cfg);
      if (!(p = strrchr(buf, DIRSEP))) {

        /* no path */
        d->path = NULL;
        d->name = strdup(buf);
      } else {

        *(p++) = '\0';
        d->path = strdup(buf);
        d->name = strdup(p);
      }
      log('D', "description_read file='%s', path='%s'", d->name, d->path);
      state = RS_LDESC;
    }

  }

  fclose(pFile);
/*  log('D', "description_read end"); */

  return list;
}


struct _desc *description_find(struct _desc *list, const char *name)
{
  struct _desc *d;

/*  log('D', "description_find name='%s', list=%p", name, list); */

  for (d=list; d; d = d->next) {
/*    log('D', "description_find comparing '%s' with '%s'", name, d->name); */
    if (!strcasecmp(name, d->name)) {
/*      log('D', "description_find found!"); */
      return d;
    }
  }
/*  log('D', "description_find not found."); */
  return NULL;
}


/* scan filespec and add all files to database */
int filespec_scan(struct _db *db, struct _cfg *cfg,
                  const char *keyname,
                  const char *path, const char *pwd, char *desc,
                  uchar entrytype, ushort level,
                  unsigned int scanopt)
{
  char *filespec;
  char buf[PATH_MAX];
  char *pBuf;
  char *descfile;
  struct _desc *listdesc, *d, *d1;
  DIR *dir;
  struct dirent *dirent;
  struct stat st;
  uchar entrytype_add = 0;
  time_t latest = 0;
  char buflatest[PATH_MAX];

/*  log('D', "filespec_scan start"); */

  if (scanopt&OKFILE_COPY)
    entrytype_add |= OKENTRY_COPY;

  /* split path from filespec */
  if (!path) {
    log('E', "internal error: filespec_scan(): no path given");
    return 0;
  }

  if (!(filespec = strrchr(path, DIRSEP))) {
    log('E', "error: path expected for filespec definition '%s'", path);
    return -1;
  }
  *(filespec++) = '\0';
  if (!*filespec)
    filespec = WC_ALLFILES;

  /* now we have the directory *path and the filespecification in *filespec */

  /* prepare buffer for full file path */
  strcpy(buf, path);
  strcat(buf, DIRSEPSTR);
  pBuf = buf + strlen(buf);

  /* read descfile from desc, if present */
/*  log('D', "filespec_scan: path='%s', filespec='%s'", path, filespec); */

  descfile = NULL;
  if (desc && *desc == '@') {
/*    log('D', "filespec_scan: have explicit descriptionfile"); */
    descfile = desc + 1;
    desc = NULL;
    if (!strchr(descfile, DIRSEP)) {
      strntcpy(pBuf, descfile, PATH_MAX - strlen(buf) - 1);
      descfile = buf;
    }
  }
  if (!descfile) {
    strcpy(pBuf, "files.bbs");
    descfile = buf;
  }
  log('D', "filespec_scan: got descriptionfile name '%s'", descfile);
  listdesc = description_read(cfg, descfile);

  if (scanopt&OKFILE_LATEST) {

    latest = 0;
    /* scan directory at <path> */
    dir = opendir(path);
    if (!dir) {
      log('E', "error: unable to scan directory '%s'", path);
    } else {
/*      log('D', "filespec_scan: scanning directory '%s'", path); */
      while ((dirent = readdir(dir)) !=  NULL) {

        if (dirent->d_type == DT_DIR || dirent->d_name[0] == '.')
          continue;

        if (fnmatch(filespec, dirent->d_name, FNM_NOESCAPE|FNM_CASEFOLD) != 0)
          continue;

        strcpy(pBuf, dirent->d_name);
        stat(buf, &st);
        if (!S_ISREG(st.st_mode))
          continue;

        if (st.st_mtime > latest) {
          latest = st.st_mtime;
          strcpy(buflatest, dirent->d_name);
        }
      }
      closedir(dir);
    }

    if (latest) {
/*      log('D', "filespec_scan: found latest file '%s'", buflatest); */
      d = description_find(listdesc, buflatest);

      if ((entrytype & OKENTRYMASK) == OKENTRY_MAGIC) {
        strcpy(pBuf, buflatest);
        db_file_add(db, keyname, pwd, buf, d ? d->desc : desc, entrytype|entrytype_add, level);
      } else
        db_file_add(db, buflatest, pwd, path, d ? d->desc : desc, entrytype|entrytype_add, level);
      if (d)
        *d->name = '\0';
    } else
      log('W', "warning: not found any file searching latest: path='%s', filespec='%s'",
           path, filespec);

  } else {

    /* scan directory at <path> */
    dir = opendir(path);
    if (!dir) {
      log('E', "error: unable to scan directory '%s'", path);
    } else {
/*      log('D', "filespec_scan: scanning directory '%s'", path); */
      while ((dirent = readdir(dir)) !=  NULL) {

        if (dirent->d_type == DT_DIR || dirent->d_name[0] == '.')
          continue;

        if (fnmatch(filespec, dirent->d_name, FNM_NOESCAPE|FNM_CASEFOLD) != 0) {
          /* log('D', "filespec_scan: no match, skip this file"); */
          continue;
        }

        strcpy(pBuf, dirent->d_name);
        /* stat(buf, &st);
        if (!S_ISREG(st.st_mode))
          continue; */

/*        log('D', "filespec_scan: found file '%s', type %d", dirent->d_name, dirent->d_type); */
        d = description_find(listdesc, dirent->d_name);

        if ((entrytype & OKENTRYMASK) == OKENTRY_MAGIC) {
          strcpy(pBuf, dirent->d_name);
          db_file_add(db, keyname, pwd, buf, d ? d->desc : desc, entrytype|entrytype_add, level);
        } else
          db_file_add(db, dirent->d_name, pwd, path, d ? d->desc : desc, entrytype|entrytype_add, level);
        if (d)
          *d->name = '\0';
      }
      closedir(dir);
    }
  }

  /* add remaining files from description file */
/*  log('D', "filespec_scan: adding remaining files from description file"); */
  for (d=listdesc; d; d = d1) {
    /* log('D', "filespec_scan: entrytype=%d, scanopt=%d", entrytype, scanopt); */
    if (entrytype == OKENTRY_FILESPEC && (scanopt&OKFILE_LATEST) == 0) {
      if (*d->name) {
/*        log('D', "filespec_scan: found filename '%s'", d->name); */
        db_file_add(db, d->name, pwd, d->path ? d->path : path, d->desc, entrytype|entrytype_add|OKENTRY_COPY, level);
      }
    }
    d1 = d->next;
    free(d->name);
    free(d->path);
    free(d->desc);
    free(d);
  }

/*  log('D', "filespec_scan end"); */
  return 0;
}


int okfile_scan(struct _db *db, struct _cfg *cfg, const char *filename, ushort level)
{
  FILE *pFile;
  char buf[LINEBUF];
  char *line;
  char *magicname, *filespec, *pwd, *desc, *option;
  char *p = NULL;
  int options;
  unsigned int opt;

/*  log('D', "okfile_scan: start"); */

  if (!filename) {
    log('E', "error: no okfile given");
    return -1;
  }

  log('I', "scanning okfile '%s'", filename);

  if (!(pFile = fopen(filename, "rt")))
  {
    log('E', "error: unable to open okfile '%s'", filename);
    return -1;
  }

  while ((line = fgets(buf, LINEBUF, pFile)) != NULL) {
    chomp(line);
    line = skipws(line);
    if (!*line || strchr(";#", *line)) continue;

/*    log('D', "okfile line='%s'", line); */

    /* get first word */
    opt = 0;
    magicname = NULL;
    pwd = NULL;
    desc = NULL;
    filespec = strtok(line, " \t");
    if (!filespec) continue;

    /* next words (may be password or options) */
    options = 1;
    while (options && (p = strtok(NULL, "\n")) != NULL) {
      switch (*p) {
        case '!':
          /* we have a password. Save it and get the rest of the line. */
          pwd = strtok(p+1, " \t");
          break;

        case '-':
           /* we have an option. Save it and get the rest of the line. */
          option = strtok(p+1, " \t");
          if (strcasecmp("copy", option) == 0) {
            opt |= OKFILE_COPY;
          } else if (strcasecmp("latest", option) == 0) {
            opt |= OKFILE_LATEST;
          } else if (strcasecmp("s", option) == 0) {
            opt |= OKFILE_SUBDIR_DESC;
          } else if (strcasecmp("t", option) == 0) {
            opt |= OKFILE_SUBDIR;
          } else {
            log('W', "warning: unrecognized option '%s' in okfile line", option);
          }
          break;

        default:
          options = 0;
          break;
      }
    }

    db_area_start(db);

    /* we have now the first word in filespec, the password (if any) in pwd, */
    /* and the rest of the line (if any) in p */
    switch (*filespec) {

      case '@':
        magicname = filespec + 1;
        if (!p) {
          log('E', "error: missing file specification for magic '%s'", magicname);
          continue;
        }
        filespec = strtok(p, " \t");
        desc = strtok(NULL, "\n");
        log('I', "magic: magicname='%s', pwd='%s', filespec='%s', desc='%s'",
                 magicname, pwd, filespec, desc);
        filespec_scan(db, cfg, magicname, filespec, pwd, desc, OKENTRY_MAGIC, level, opt);
        break;

      case '$':
        magicname = filespec + 1;
        if (!p) {
          log('E', "error: missing command specification for external command '%s'", magicname);
          continue;
        }
        filespec = p;
        log('I', "external cmd: magicname='%s', pwd='%s', cmdline='%s'",
                 magicname, pwd, filespec);
        db_file_add(db, magicname, pwd, filespec, "[external program called]", OKENTRY_CMD, level);
        break;

      default:
        desc = p;
        log('I', "filespec: filespec='%s', pwd='%s', desc='%s'",
                 filespec, pwd, desc); 
        filespec_scan(db, cfg, NULL, filespec, pwd, desc, OKENTRY_FILESPEC, level, opt);
    }

    db_area_end(db);

  }
  fclose(pFile);
/*  log('D', "okfile_scan: end"); */
  return 0;
}


int rp_findfiles(struct _db *db,
                 const char *filespec, const char *pwd, ushort level,
                 CB_FINDFILE pCBFunc, void *pData)
{
  size_t prefixlen = strlen(filespec);
  long idx;
  struct _db_info info;
  int numfound = 0;
  char *filename = (char *) malloc(PATH_MAX);
  char *filepwd  = (char *) malloc(PWDLEN);
  char *filepath = (char *) malloc(PATH_MAX);
  char *filedesc = (char *) malloc(DESCBUF);

  log('D', "rp_findfiles start");

  if (!(filename && filepwd && filepath && filedesc)) {
    log('E', "error: unable to allocate memory in rp_findfiles");
    exit(4);
  }

  prefixlen = strcspn(filespec, "*?");

  log('I', "searching pattern='%s' password='%s' level=%hu", filespec, pwd, level);

  for (idx = db_idx_search(db, filespec, prefixlen);
       idx != -1 && idx < db->idxitems;
       idx++) {
/*    log('D', "rp_findfiles: found entry at record %ld", idx); */

    /* get file info entry */
/*    log('D', "rp_findfiles: call db_idx_info_get"); */
    if (db_idx_info_get(db, idx, &info) != 0)
      break;

    /* get filename */
/*    log('D', "rp_findfiles: call db_heapstr_get for filename"); */
    db_heapstr_get(db->pNameIdx, info.ofsName, filename, PATH_MAX);

    /* prefix still matches? */
/*    log('D', "rp_findfiles: comparing '%s' and '%s', len=%d", filespec, filename, prefixlen); */
    if (strncasecmp(filespec, filename, prefixlen))
      break;

    if (info.usLevel > level) {
      log('I', "caller's level too low: filelevel=%hu, callerlevel=%hu",
               info.usLevel, level);
      continue;
    }

/*    log('D', "rp_findfiles: matching filespec='%s' with '%s'", filespec, filename); */
    if (fnmatch(filespec, filename, FNM_NOESCAPE|FNM_CASEFOLD) == 0) {
      log('I', "matching entry: %s", filename);

/*      log('D', "matches: '%s' and '%s'", filespec, filename); */

      if (info.ofsPwd != -1) {
        if (!pwd) {
          log('W', "PASSWORD REQUIRED, but none given info.ofsPwd=%ld", info.ofsPwd);
          continue;
        }
        db_heapstr_get(db->pPwdIdx, info.ofsPwd, filepwd, PWDLEN);
        if (!strcasecmp(pwd, filepwd)) {
          log('D', "PASSWORD MATCHES: '%s'", filepwd);
        } else {
          log('W', "PASSWORD MISMATCH: pwd='%s', filepwd='%s'", pwd, filepwd);
          continue;
        }
      }

      db_heapstr_get(db->pPathIdx, info.ofsPath, filepath, PATH_MAX);
/*      log('D', "PATH: '%s'", filepath); */

      db_heapstr_get(db->pDescIdx, info.ofsDesc, filedesc, DESCBUF);
/*      log('D', "DESCRIPTION: '%s'", filedesc); */

      numfound += pCBFunc(pData, filename, filepath, filedesc, info.ucEntryType);
    }
  }
  if (!numfound)
    pCBFunc(pData, filespec, NULL, NULL, OKENTRY_NOTFOUND);

  log('D', "rp_findfiles end");

  free(filename);
  free(filepwd);
  free(filepath);
  free(filedesc);

  return 0;
}


int fprint_wordwrap(const char *s, int lmargin, int rmargin, FILE *stream, char nl)
{
  int width = rmargin - lmargin + 1;
  char buf[256];
  char *d, *line;
  int idx, slen;
  int lines = 0;

  memset(buf, ' ', 256);
  d = line = buf + lmargin;

  while (s && *s) {
    if (!(slen = strcspn(s, "\n")))
      slen = strlen(s);
    while (slen > 0) {
      idx = slen;
      if (idx > width) {
        idx = width;
        while (idx && !isspace(s[idx])) --idx;
        if (!idx) idx = width;
      }
      memcpy(d, s, idx);
      d[idx] = '\0';
      if (isspace(s[idx])) ++idx;
      s += idx;
      slen -= idx;
      fputs(line, stream);
      putc(nl, stream);
      ++lines;
      line = buf;
    }
    if (*s == '\n') ++s;
  }
  if (!lines)
    putc(nl, stream);

  return lines;
}


int find_file_cb(void *pData, const char *name,
                  const char *path, const char *desc, char type)
{
  struct stat st;
  char fname[PATH_MAX];
  char fdate[11];       /* MM.DD.YYYY */
  char *p;
  int result = 0;
  char *s;

  pData=pData; /* unused */

  s = (type&OKENTRY_COPY) ? "[COPY TO TEMP]" : "";

  switch (type & OKENTRYMASK) {
    case OKENTRY_FILESPEC:
      snprintf(fname, PATH_MAX, "%s%c%s", path, DIRSEP, name);
      /* printf("NAME: '%s'\n  FILENAME: '%s'\n  DESC: '%s'\n",
             name, fname, desc); */
      if (!stat(fname, &st) && S_ISREG(st.st_mode)) {
        printf("SEND FILE '%s' %s\n", fname, s);
        strftime(fdate, 11, "%d.%m.%Y", localtime(&st.st_mtime));
        printf("%-12s %9lu %s ", name, st.st_size, fdate);
        fprint_wordwrap(desc, 34, 78, stdout, '\n');
        result = 1;
      }
      break;
    case OKENTRY_MAGIC:
      /* printf("NAME: '%s'\n  FILENAME: '%s'\n  DESC: '%s'\n", name, path, desc); */
      if (!(p = strrchr(path, DIRSEP)))
        break;
      if (!stat(path, &st) && S_ISREG(st.st_mode)) {
        printf("SEND FILE '%s' %s\n", path, s);
        strftime(fdate, 11, "%d.%m.%Y", localtime(&st.st_mtime));
        printf("%-12s %9lu %s ", ++p, st.st_size, fdate);
        fprint_wordwrap(desc, 34, 78, stdout, '\n');
        result = 1;
      }
      break;

    case OKENTRY_CMD:
      /* printf("NAME: '%s'\n  EXECUTE: '%s'\n", name, path); */
      printf("EXECUTE: '%s'\n", path);
      printf("%-12s                      ", name);
      fprint_wordwrap(desc, 34, 78, stdout, '\n');
      result = 1;
      break;

    case OKENTRY_NOTFOUND:
/*      printf("NOTFOUND: '%s'\n", name); */
      printf("%-12s                      ", name);
      fprint_wordwrap("File not found or password error", 34, 78, stdout, '\n');
      break;

    default:
/*      printf("ENTRY TYPE ERROR, type=%d\n", (int) type); */
      printf("%-12s                      ", name);
      fprint_wordwrap("internal program error processing this request", 34, 78, stdout, '\n');
      break;
  }

  return result;
}


int srif_file_cb(void *pData, const char *name,
                  const char *path, const char *desc, char type)
{
  struct stat st;
  char fname[PATH_MAX];
  char tmpfname[PATH_MAX];
  char fdate[11];       /* MM.DD.YYYY */
  int rc;
  int result = 0;
  struct _cfg *cfg = (struct _cfg *) pData;
  int copytotemp = (type&OKENTRY_COPY);
  char prefix = '+';

  copytotemp = copytotemp;

  type &= OKENTRYMASK;

  switch (type) {
    case OKENTRY_MAGIC:
      if (!(name = strrchr(path, DIRSEP)))
        break;
      name++;
      /* fallthrough */

    case OKENTRY_FILESPEC:
      if (type==OKENTRY_MAGIC)
        strntcpy(fname, path, PATH_MAX);
      else
        snprintf(fname, PATH_MAX, "%s%c%s", path, DIRSEP, name);

      if (!stat(fname, &st)) {
        if (S_ISREG(st.st_mode)) {
          /* check time, size and number of files */

          log('D', "srif_file_cb: remaining_seconds=%d remaining_files=%d remaining_bytes=%d",
                   cfg->remaining_seconds, cfg->remaining_files, cfg->remaining_bytes);
          log('D', "srif_file_cb: filetime=%d filesize=%d baud=%d",
                   st.st_size / cfg->srif_baud, st.st_size, cfg->srif_baud);

          if (cfg->remaining_seconds < st.st_size / cfg->srif_baud) {
            desc = "Session time limit exceeded or no time to next non-request event";
            log('I', "session time limit exceeded: remaining_seconds=%d < needed=%d",
              cfg->remaining_seconds, st.st_size / cfg->srif_baud);
          } else if (cfg->remaining_files < 1) {
            desc = "File limit exceeded";
            log('I', "file limit exceeded: remaining_files=%d", cfg->remaining_files);
          } else if (cfg->remaining_bytes < st.st_size) {
            desc = "File size limit exceeded";
            log('I', "file size exceeded: remaining_bytes=%d < filesize=%d",
                      cfg->remaining_bytes, st.st_size);
          } else {
            /* all ok: copy file to temp, if necessary */
            if (copytotemp) {
              snprintf(tmpfname, PATH_MAX, "%s%c%s", get_tempdir(), DIRSEP, name);
              log('I', "copy file %s to temp %s", fname, tmpfname);
              if (file_copy(fname, tmpfname, 0) == -1) {
                log('W', "unable to copy file '%s' to temp '%s', sending original",
                       fname, tmpfname);
              } else {
                prefix = '-';
                strcpy(fname, tmpfname);
              }
            }
            log('I', "sending %s", fname);
            result = 1;
            putc(prefix, cfg->pResponselist);
            fputs(fname, cfg->pResponselist);
            putc('\n', cfg->pResponselist);
            --cfg->remaining_files;
            cfg->remaining_bytes -= st.st_size;
            cfg->remaining_seconds -= st.st_size / cfg->srif_baud;
          }
          log('D', "send file '%s' with desc='%s'", fname, desc);
          strftime(fdate, 11, "%d.%m.%Y", localtime(&st.st_mtime));
          fprintf(cfg->pPkt, "%-12s %9lu %s ", name, st.st_size, fdate);
        } else {
          log('W', "entry '%s' is not a regular file - skipping silently", fname);
          return result;
        }
      } else {
        log('W', "file in index but not on disk: '%s'", fname);
        fprintf(cfg->pPkt, "%-12s                      ", name);
        desc = "internal error: local file not accessible";
      }

      break;

    case OKENTRY_CMD:
      snprintf(fname, PATH_MAX, path, (int) cfg->srif_akamatch->zone,
                                      (int) cfg->srif_akamatch->net,
                                      (int) cfg->srif_akamatch->node,
                                      (int) cfg->srif_akamatch->point);
      log('I', "Execute cmd='%s'", fname);
      rc = system(fname);
      log('I', "Execute returned rc=%d", rc);
      result = 1;
      fprintf(cfg->pPkt, "%-12s                      ", name);
      break;

    case OKENTRY_NOTFOUND:
      fprintf(cfg->pPkt, "%-12s                      ", name);
      desc = "File not found or password error";
      break;

    default:
      fprintf(cfg->pPkt, "%-12s                      ", name);
      desc = "internal program error processing this request";
      break;
  }

  fprint_wordwrap(desc, 34, 78, cfg->pPkt, '\r');

  return result;
}



int find_file(struct _cfg *cfg, const char *wildcard, const char *pwd)
{
  struct _db *db;
  int rc;

  if ((db=db_open_read(cfg))==NULL)
    return -1;

  rc = rp_findfiles(db, wildcard, pwd, cfg->level, find_file_cb, NULL);

  db_close(db);
  return rc;
}


int srif_read(struct _cfg *cfg, const char *filename)
{
  char buf[LINEBUF];
  FILE *pFile;
  char *line, *kw, *value, *p;
  int rc = -1;

  cfg->srif_sysop = NULL;
  cfg->srif_aka = NULL;
  cfg->srif_akamatch = NULL;
  cfg->srif_baud = 0;
  cfg->srif_time = -1;
  cfg->srif_requestlist = NULL;
  cfg->srif_responselist = NULL;
  cfg->srif_userinfo = NULL;
  cfg->srif_password[0] = '\0';

  if (!filename || !(pFile = fopen(filename, "rt"))) {
    log('E', "unable to open srif file '%s'", filename);
    return -1;
  }

  while ((line = fgets(buf, LINEBUF, pFile)) != NULL) {
    line = skipws(line);

    kw = strtok(line, " \t");
    if (!kw) continue;

    value = strtok(NULL, "\n");
    if (!value) continue;
    value = skipws(value);
    p = value + strlen(value)-1;
    while (p > value && isspace(*p))
      *(p--) = '\0';
    if (!*value) continue;

    log('D', "srif_read: kw='%s', value='%s'", kw, value);

    if (!strcasecmp("sysop", kw)) {
      cfg->srif_sysop = strdup(value);
    } else if (!strcasecmp("aka", kw)) {
      struct _fidonet_addrlist *a = (struct _fidonet_addrlist *)malloc(sizeof(struct _fidonet_addrlist));
      a->next = cfg->srif_aka;
      cfg->srif_aka = a;
      fidonet_str2addr(value, &a->addr);
    } else if (!strcasecmp("baud", kw)) {
      cfg->srif_baud = atol(value);
    } else if (!strcasecmp("time", kw)) {
      cfg->srif_time = atol(value);
    } else if (!strcasecmp("requestlist", kw)) {
      cfg->srif_requestlist = strdup(value);
    } else if (!strcasecmp("responselist", kw)) {
      cfg->srif_responselist = strdup(value);
    } else if (!strcasecmp("remotestatus", kw)) {
      if (!strcasecmp("protected", value))
        cfg->srif_userinfo = &cfg->protected;
    } else if (!strcasecmp("systemstatus", kw)) {
      if (!cfg->srif_userinfo && !strcasecmp("listed", value))
        cfg->srif_userinfo = &cfg->listed;
      else
        cfg->srif_userinfo = &cfg->unlisted;
    } else if (!strcasecmp("password", kw)) {
        strntcpy(cfg->srif_password, value, 9);
    }
  }

  /* sanity tests */
  if (!cfg->srif_sysop)
    cfg->srif_sysop = strdup("Sysop");
  if (cfg->srif_baud == 0)
    cfg->srif_baud = 300;

  /* do we have sufficient information from srif to process request? */
  if (cfg->srif_userinfo && cfg->srif_requestlist && cfg->srif_responselist) {
    rc = 0;

    fidonet_akamatch(cfg->myaddr, cfg->srif_aka,
                     &cfg->myaddrmatch, &cfg->srif_akamatch);
    if (!cfg->srif_akamatch || !cfg->myaddrmatch) {
      log('E', "error: insufficient address information - unable to create response");
      rc = 1;
    }

    cfg->level = cfg->srif_userinfo->level;
    cfg->remaining_seconds = ULONG_MAX;
    if (cfg->srif_time != -1)
      cfg->remaining_seconds = cfg->srif_time * 60;
    if (cfg->remaining_seconds > cfg->srif_userinfo->maxtime)
      cfg->remaining_seconds = cfg->srif_userinfo->maxtime;
    cfg->remaining_files = cfg->srif_userinfo->maxfiles;
    cfg->remaining_bytes = cfg->srif_userinfo->maxbytes;

    log('I', "%s (%s) calling, level=%d", cfg->srif_sysop,
         fidonet_addr2str4d(buf, cfg->srif_akamatch), cfg->level);
    log('I', "available seconds=%d, files=%d, bytes=%d",
         cfg->remaining_seconds, cfg->remaining_files, cfg->remaining_bytes);
  }

  return rc;
}



int request_srif(struct _cfg *cfg, const char *srifname)
{
  FILE *pRequestlist;
  char buf[LINEBUF];
  char *line, *req, *pwd;
  struct _db *db;
  int rc;
  char fbuf[PATH_MAX];
  char *stemp;
  unsigned long pktnr;
  int pktfd;
  struct _pkthdr pkthdr;
  struct _msghdr msghdr;

  /* open database */
  if ((db=db_open_read(cfg))==NULL) {
    return -1;
  }

  /* read srifname */
  if (srif_read(cfg, srifname) != 0) {
    db_close(db);
    log('E', "error: unable to get sufficient data from srif file '%s'", srifname);
    return -1;
  }

  /* open requestlist */
  if (!(pRequestlist = fopen(cfg->srif_requestlist, "rt"))) {
    db_close(db);
    log('E', "error: unable to open srif request list '%s'", cfg->srif_requestlist);
    return -1;
  }

  /* create responselist */
  if (!(cfg->pResponselist = fopen(cfg->srif_responselist, "wt"))) {
    db_close(db);
    fclose(pRequestlist);
    log('E', "error: unable to open srif response list '%s'", cfg->srif_responselist);
    return -1;
  }

  stemp = get_tempdir();

  pktnr = GenMsgId();
  do {
    char buf1[9];
    snprintf(buf1, 9, "%08lu", pktnr);
    snprintf(fbuf, PATH_MAX, "%s%c%s.pkt", stemp, DIRSEP, buf1);
    log('D', "request_srif: trying to create pkt file '%s'", fbuf);
    pktfd = open(fbuf, O_EXCL|O_CREAT|O_TRUNC|O_WRONLY|O_BINARY,
                       S_IRUSR|S_IWUSR);
    ++pktnr;
  } while (pktfd == -1 && pktnr != 0);

  if (pktfd == -1 || !(cfg->pPkt = fdopen(pktfd, "wb"))) {
    db_close(db);
    fclose(cfg->pResponselist);
    unlink(cfg->srif_responselist);
    fclose(pRequestlist);
    log('E', "error: unable to create response pkt file '%s'", fbuf);
    return -1;
  }

  pkthdr.destAddr = *cfg->srif_akamatch;
  pkthdr.origAddr = *cfg->myaddrmatch;
  pkthdr.hiProductCode = 0;
  pkthdr.loProductCode = 254;
  pkthdr.majorProductRev = 1;
  pkthdr.minorProductRev = 0;
  pkthdr.created = time(NULL);
  pkthdr.capabilityWord = 1;
  pkthdr.prodData = 0;
  strcpy(pkthdr.password, cfg->srif_password);

  pkt_writehdr(cfg->pPkt, &pkthdr);


  msghdr.destAddr = *cfg->srif_akamatch;
  msghdr.origAddr = *cfg->myaddrmatch;
  msghdr.attribute = MSGPRIVATE|MSGLOCAL;
  msghdr.datetime[0] = '\0';
  msghdr.timestamp = time(NULL);

  strntcpy(msghdr.toUserName, cfg->srif_sysop, 36);
  strntcpy(msghdr.fromUserName, "Request Processor", 36);
  strntcpy(msghdr.subject, "Response to file request", 72);

  pkt_writemsg(cfg->pPkt, &msghdr, NULL);

  putc('-', cfg->pResponselist);
  fputs(fbuf, cfg->pResponselist);
  putc('\n', cfg->pResponselist);

  /* for each line in request list: perform request */
  while ((line = fgets(buf, LINEBUF, pRequestlist)) != NULL) {
    chomp(line);

    log('D', "request_srif: req='%s'", line);

    req = strtok(line, " \t");
    if (!req) continue;
    pwd = strtok(NULL, " \t");
    if (pwd) {
      if (*pwd == '!')
        ++pwd;
      else
        pwd = NULL;
    }
    rc = rp_findfiles(db, req, pwd, cfg->level, srif_file_cb, cfg);
  }

  /* finished */
  fwrite("\000", 1, 1, cfg->pPkt);
  fclose(cfg->pResponselist);
  fclose(pRequestlist);
  pkt_close(cfg->pPkt);

  db_close(db);

  return 0;
}

/* -------------------------------------------------------------------------- */
/*
Build index:
- files containing the index:
  file_path.idx   all paths to the files
  file_desc.idx   file descriptions
  file_name.idx   file names
  file_pwd.idx    file passwords, if configured
  file_info.idx   offsets for path+descriptions+name+password for every files
  file_idx.idx    all offsets in file_info.idx, sorted by file name

- for every filespec from okfile:
  - scan files in directory
  - read file list (files.bbs) if present, merge with scanned files
  - add file directory to path list
  - add file description to description list
  - add file name to file list
  - write file list to disk:
      file names to file_name.idx,
      path names to file_path.idx,
      descriptions to file_desc.idx,
      password to file_pwd.idx,
      offsets and entry types to file_info.idx
  - keep offsets to file_info.idx and corresponding file names in memory
- sort list of offsets by file name
- write sorted offsets to file_idx.idx
*/
int create_index(struct _cfg *cfg)
{
  struct _db *db;
  unsigned int i;

  if ((db = db_create(cfg)) == NULL)
    return -1;

  for (i=0; i < cfg->numokfiles; i++) {
    okfile_scan(db, cfg, cfg->okfile[i].name, cfg->okfile[i].level);
  }
  db_sort_idx(db);
  db_write_idx(db);
  db_close(db);
  return 0;
}



int conf_userinfo(struct _db_userinfo *pUserInfo, char *opt)
{
/* opt = <lvl>:<maxfiles>:<maxbytes>:<maxtime>:<template> */
  int num;


  pUserInfo->template = strdup(opt);
  num = sscanf(opt, "%hu:%lu:%lu:%lu:%s", &pUserInfo->level,
                                          &pUserInfo->maxfiles,
                                          &pUserInfo->maxbytes,
                                          &pUserInfo->maxtime,
                                          pUserInfo->template);
  if (num != 5) {
    free(pUserInfo->template);
    pUserInfo->template = NULL;
    log('E', "error: syntax error in user info: got='%s', expected='<lvl>:<maxfiles>:<maxbytes>:<maxtime>:<template>'", opt);
    return 0;
  }

  if (pUserInfo->maxtime < pUserInfo->maxtime * 60) /* handle overflow */
    pUserInfo->maxtime = pUserInfo->maxtime * 60;

  if (pUserInfo->maxfiles == 0) pUserInfo->maxfiles = ULONG_MAX;
  if (pUserInfo->maxbytes == 0) pUserInfo->maxbytes = ULONG_MAX;
  if (pUserInfo->maxtime == 0)  pUserInfo->maxtime  = ULONG_MAX;

  log('D', "conf_userinfo: scanned %d userinfo parameters: level=%hu, " \
           "maxfiles=%lu, maxbytes=%lu, maxtime=%lu, template=%s",
           num,
           pUserInfo->level, pUserInfo->maxfiles, pUserInfo->maxbytes,
           pUserInfo->maxtime, pUserInfo->template);

  return num;
}


void usage(void)
{
  printf("\
rproc " VERSION " - a tiny Fidonet srif file request processor\n\
(c) 2004-2012 by Alexander Woick, 2:244/1351\n\
rproc [<options>] <command>\n\
\n\
commands:\n\
  <path-to-srif>        perform request according to srif file\n\
  srif <path-to-srif>   perform request according to srif file\n\
  index                 create request index\n\
  dump                  dump the index\n\
  find <filespec> <password> find file(s)\n\
\n\
options:\n\
  --log <logfile>       logfile\n\
  --quiet               no visible logging output\n\
  --verbose             verbose logging\n\
  --help                print help\n\
  --reqidxdir <path>    path to request index\n\
\n\
  --level <lvl>         access level for --okfile\n\
  --okfile <file>       okfile.lst for indexing (can be repeated)\n\
\n\
  --unlisted <lvl>:<maxfiles>:<maxbytes>:<maxtime>:<template>\n\
  --listed <lvl>:<maxfiles>:<maxbytes>:<maxtime>:<template>\n\
  --protected <lvl>:<maxfiles>:<maxbytes>:<maxtime>:<template>\n\
\n\
  --map <from>=<to>     map path components during indexing\n\
  --addr <addr>[,<addr>, ...] supply origin addresses for responses\n\
\n\
remarks: \n\
  - template not implemented yet. Response is hardcoded.\n\
");

  exit(1);
}

void short_usage(void)
{
  fprintf(stderr, "usage: rproc [-options] <command>\n");
  fprintf(stderr, "       rproc --help  for more information\n");
  exit(1);
}


/*
srif file:
sysop <remote sysopname>
aka <remote z:f/n.p@d>
baud <rate>
time <time-to-next-nonrequest-event>
requestlist <filename of request list>
responselist <filename of response list>
             format: one filename per line, first char:
               = erase file after successful transmission
               + simply send, no erase
               - erase file in any case after session
remotestatus protected|unprotected
systemstatus listed|unlisted
# optional:
Site <remote siteinfo>
Location <remote location>
Phone <remote phone>
Mailer <remote mailer>
PORT <port>
Password <session password>

*/

int main(int argc, char *argv[])
{
  int rc;
  int option_index;
  int c;
  int i;
  unsigned ui;
  int cmdc;
  char **cmdv;
  char *p;
  struct _fidonet_addrlist *ap, *ap1;
  struct _cfg *cfg = (struct _cfg *) malloc(sizeof(struct _cfg));

  static struct option long_options[] = {
    { "reqidxdir", 1, 0, 'r'},  /* directory of index files */
    { "level",     1, 0, 'L'},  /* access level for --okfile */
    { "okfile",    1, 0, 'o'},  /* okfile.lst for indexing (can be repeated) */
    { "unlisted",  1, 0, 'u'},  /* userinfo for unlisted systems */
    { "listed",    1, 0, 'l'},  /* userinfo for listed systems */
    { "protected", 1, 0, 'p'},  /* userinfo for protected systems */
    { "map",       1, 0, 'm'},  /* map in paths */
    { "addr",      1, 0, 'a'},  /* address list */

    { "log",       1, 0, 'g'},  /* logfile */
    { "quiet",     0, 0, 'q'},  /* no visible output */
    { "verbose",   0, 0, 'v'},  /* More verbose */
    { "help",      0, 0, 'h'},  /* Help */
    { 0,           0, 0, 0  }
  };

  if (!cfg) {
    fprintf(stderr, "Unable to allocate memory for configuration data\n");
    exit(4);
  }

  cfg->reqidxdir = strdup(".");
  cfg->level = 0;
  cfg->okfile = NULL;
  cfg->numokfiles = 0;
  cfg->mapfrom = cfg->mapto = NULL;
  cfg->nummap = 0;
  cfg->myaddr = NULL;
  cfg->myaddrmatch = NULL;

  cfg->srif_sysop = NULL;
  cfg->srif_aka = NULL;
  cfg->srif_requestlist = NULL;
  cfg->srif_responselist = NULL;

  cfg->got_unlisted = 0;
  cfg->got_listed = 0;
  cfg->got_protected = 0;
  memset(&cfg->unlisted, 0, sizeof(cfg->unlisted));
  memset(&cfg->listed, 0, sizeof(cfg->listed));
  memset(&cfg->protected, 0, sizeof(cfg->protected));

  while ((c = getopt_long(argc, argv, "r:L:o:u:l:p:vh",
                          long_options, &option_index)) != -1) {
    switch (c) {

      case 'r':
        free(cfg->reqidxdir);
        cfg->reqidxdir = strdup(optarg);
        break;

      case 'L':
        cfg->level = (unsigned short) atol(optarg);
        break;

      case 'o':
        ++cfg->numokfiles;
        cfg->okfile = (struct _okfile *) realloc(cfg->okfile,
                                    sizeof(struct _okfile) * cfg->numokfiles);
        cfg->okfile[cfg->numokfiles-1].name = strdup(optarg);
        cfg->okfile[cfg->numokfiles-1].level = cfg->level;
        break;

      case 'u':
        cfg->got_unlisted = conf_userinfo(&cfg->unlisted, optarg);
        break;

      case 'l':
        cfg->got_listed = conf_userinfo(&cfg->listed, optarg);
        break;

      case 'p':
        cfg->got_protected = conf_userinfo(&cfg->protected, optarg);
        break;

      case 'm':
        i = strcspn(optarg, "=");
        if (!i) {
          fprintf(stderr, "map option error: expected <from>=<to>\n");
          short_usage();
        }
        ++cfg->nummap;
        cfg->mapfrom = (char**) realloc(cfg->mapfrom, sizeof(char *) * cfg->nummap);
        cfg->mapto = (char**) realloc(cfg->mapto, sizeof(char *) * cfg->nummap);
        cfg->mapfrom[cfg->nummap-1] = (char *) malloc(++i);
        strntcpy(cfg->mapfrom[cfg->nummap-1], optarg, i);
        cfg->mapto[cfg->nummap-1] = strdup(optarg+i);
        /* printf("mapfrom='%s', mapto='%s'\n", cfg->mapfrom[cfg->nummap-1], cfg->mapto[cfg->nummap-1]); */
        break;

      case 'a':
        /* printf("main: parsing '--addr=%s'", optarg); */
        if (!(p = strtok(optarg, ","))) p = optarg;
        while (p) {
          struct _fidonet_addrlist *a = (struct _fidonet_addrlist *) malloc(sizeof(struct _fidonet_addrlist));
          a->next = cfg->myaddr;
          cfg->myaddr = a;
          fidonet_str2addr(p, &a->addr);
          /* printf("addr=%s\n", fidonet_addr2str4d(buf, &a->addr)); */
          p = strtok(NULL, ",");
        }
        break;

      /***** Common options *****/
      case 'q':
        logquiet = 1;
        break;
      case 'g':
        logfile = strdup(optarg);
        break;
      case 'v':
        loglevel = "XEWITD";
        break;
      case 'h':
        usage();
        break;
      default:
        short_usage();
        break;
    }
  }

  if (!cfg->got_listed) cfg->listed = cfg->unlisted;
  if (!cfg->got_protected) cfg->protected = cfg->listed;

  if (optind >= argc) {
    fprintf(stderr, "rproc: expecting command or srif file\n");
    short_usage();
  }
  cmdc = argc - optind;
  cmdv = &argv[optind];

  if (!strcasecmp("index", cmdv[0])) {

    rc = create_index(cfg);

  } else if (!strcasecmp("dump", cmdv[0])) {

    rc = db_dump(cfg);

  } else if (!strcasecmp("find", cmdv[0])) {
    if (cmdc < 2) {
      fprintf(stderr, "rproc: expecting wildcard after 'find' command\n");
      short_usage();
    }
    rc = find_file(cfg, cmdv[1], cmdv[2]);

  } else if (!strcasecmp("srif", cmdv[0])) {
    if (cmdc < 2) {
      fprintf(stderr, "rproc: expecting srif file after 'srif' command\n");
      short_usage();
    }
    rc = request_srif(cfg, cmdv[1]);

  } else {

    rc = request_srif(cfg, cmdv[0]);

  }

  free(cfg->reqidxdir);

  for (ui = 0; ui < cfg->numokfiles; ui++) {
    free(cfg->okfile[ui].name);
  }
  free(cfg->okfile);

  for (ui = 0; ui < cfg->nummap; ui++) {
    free(cfg->mapfrom[ui]);
    free(cfg->mapto[ui]);
  }
  free(cfg->mapfrom);
  free(cfg->mapto);

  if (cfg->got_unlisted)  free(cfg->unlisted.template);
  if (cfg->got_listed)    free(cfg->listed.template);
  if (cfg->got_protected) free(cfg->protected.template);

  for (ap = cfg->myaddr; ap; ap = ap1) {
    ap1 = ap->next;
    free(ap);
  }
  for (ap = cfg->srif_aka; ap; ap = ap1) {
    ap1 = ap->next;
    free(ap);
  }
  free(cfg->srif_sysop);
  free(cfg->srif_requestlist);
  free(cfg->srif_responselist);

  free(cfg);
  return rc;
}


/*
TODO:
  - -s option to include subdirectories in okfile directory definition with files.bbs
  - -t option to include subdirectories in okfile directory definition without files.bbs
  - template response file
*/
